/* ==========================================================================
 * Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * ==========================================================================
 */
#include <stddef.h>		/* size_t */

#include <string.h>		/* memcpy(3) strncasecmp(3) */

#include <errno.h>		/* ENAMETOOLONG */

#include <assert.h>		/* assert */

#include <time.h>		/* time(2) */

#include <sys/param.h>		/* MIN */
#include <sys/uio.h>		/* struct iovec */

#include <pthread.h>		/* PTHREAD_MUTEX_INITIALIZER pthread_mutex_t
				 * */

#include <evnet/thread.h>	/* thread_run() */

#include "log.h"
#include "zmbuf.h"
#include "znode.h"
#include "bdb.h"
#include "edit.h"


#ifndef arraylen
#define arraylen(a)	(sizeof (a) / sizeof (a)[0])
#endif


static const DBT dbt_initializer;


static struct {
	pthread_mutex_t mutex;
	pthread_cond_t gate;
	SIMPLEQ_HEAD(, zmbuf) queue;
} edits = {
	.mutex	= PTHREAD_MUTEX_INITIALIZER,
	.gate	= PTHREAD_COND_INITIALIZER,
	.queue	= SIMPLEQ_HEAD_INITIALIZER(edits.queue),
};



static size_t edit_rdata_any(unsigned char *buf, size_t bufsiz, struct ZonedRR *rr)  {
	unsigned long repid	= time(0);
	char *host		= (char *)rr->any.host.buf;
	int hlen		= (int)rr->any.host.size;
	unsigned long ttl	= rr->any.ttl;
	char *rtype		= (char *)rr->any.rtype.buf;
	int rtlen		= (int)rr->any.rtype.size;
	char *rdata		= (char *)rr->any.rdata.buf;
	int rdlen		= (int)rr->any.rdata.size;
	int num;

	num	= snprintf((char *)buf, bufsiz, "%lu %.*s %lu %.*s %.*s", repid, hlen, host, ttl, rtlen, rtype, rdlen, rdata);

	return (num < 0)? 0 : num;
} /* edit_rdata_any() */


static size_t edit_rdata_a(unsigned char *buf, size_t bufsiz, struct ZonedRR *rr)  {
	unsigned long repid	= time(0);
	char *host		= (char *)rr->a.host.buf;
	int hlen		= (int)rr->a.host.size;
	unsigned long ttl	= rr->a.ttl;
	char *address		= (char *)rr->a.address.buf;
	int alen		= (int)rr->a.address.size;
	int num;

	num	= snprintf((char *)buf, bufsiz, "%lu %.*s %lu A %.*s", repid, hlen, host, ttl, alen, address);

	return (num < 0)? 0 : num;
} /* edit_rdata_a() */


static size_t edit_rdata_srv(unsigned char *buf, size_t bufsiz, struct ZonedRR *rr)  {
	unsigned long repid	= time(0);
	char *host		= (char *)rr->srv.host.buf;
	int hlen		= (int)rr->srv.host.size;
	unsigned long ttl	= rr->srv.ttl;
	unsigned long priority	= rr->srv.priority;
	unsigned long weight	= rr->srv.weight;
	unsigned short port	= rr->srv.port;
	char *target		= (char *)rr->srv.target.buf;
	int tlen		= (int)rr->srv.target.size;
	int num;

	num	= snprintf((char *)buf, bufsiz, "%lu %.*s %lu SRV %lu %lu %hu %.*s", repid, hlen, host, ttl, priority, weight, port, tlen, target);

	return (num < 0)? 0 : num;
} /* edit_rdata_srv() */


static size_t edit_rdata(void *buf, size_t bufsiz, struct ZonedRR *rr)  {
	static const typeof (&edit_rdata_any) rdata[]	= {
		[ZonedRR_PR_any]	= &edit_rdata_any,
		[ZonedRR_PR_a]		= &edit_rdata_a,
		[ZonedRR_PR_srv]	= &edit_rdata_srv,
	};
	if (rr->present < 0 || rr->present >= arraylen(rdata) || 0 == rdata[rr->present]) {
		error("unknown RR type: %d", rr->present);

		return 0;
	}

	return rdata[rr->present](buf, bufsiz, rr);
} /* edit_rdata() */


static size_t edit_rzone(unsigned char *dstbuf, size_t dstsiz, unsigned char *srcbuf, size_t srcsiz) {
	struct { unsigned char *beg, *pos, *end; } dst, src;

	dst.pos	= dstbuf;
	dst.end	= dstbuf + dstsiz;

	src.pos	= srcbuf + srcsiz;
	src.beg	= srcbuf;

	while (dst.pos < dst.end && src.pos > src.beg) {
		src.pos--;

		*dst.pos	= *src.pos;

		dst.pos++;
	}

	return dst.pos - dstbuf;
} /* edit_rzone() */


static int edit_write_zone_set(struct zmbuf *mb, struct bdb *bdb, DB_TXN *txn, const char *logid, struct ZonedRR *rr) {
	unsigned char zone[256];
	size_t zlen;
	DBT key		= dbt_initializer;
	DBT data	= dbt_initializer;
	int dberr;

	zlen	= edit_rzone(zone, sizeof zone, mb->msg->edit.zone.buf, mb->msg->edit.zone.size);

	if (zlen >= sizeof zone) {
		error("(%s) zone name too long: `%.*s...'", logid, (int)MIN(255, mb->msg->edit.zone.size), mb->msg->edit.zone.buf);

		return ENAMETOOLONG;
	}

	zone[zlen]	= '\0';

	key.data	= zone;
	key.size	= zlen;

	dberr	= bdb->db.zone->put(bdb->db.zone, txn, &key, &data, DB_NOOVERWRITE);

	switch (dberr) {
	case DB_KEYEXIST:
		/* FALL THROUGH */
	case 0:
		return 0;
	default:
		return dberr;
	}

	/* NOT REACHED */
} /* edit_write_zone_set() */


static int edit_write_zone_del(struct zmbuf *mb, struct bdb *bdb, DB_TXN *txn, const char *logid, struct ZonedRR *rr) {
	DBT key		= dbt_initializer;
	int dberr;

	key.data	= mb->msg->edit.zone.buf;
	key.size	= mb->msg->edit.zone.size;

	dberr	= bdb->db.zone->del(bdb->db.zone, txn, &key, 0);

	switch (dberr) {
	case DB_NOTFOUND:
		/* FALL THROUGH */
	case 0:
		return 0;
	default:
		return dberr;
	}

	/* NOT REACHED */
} /* edit_write_zone_del() */


/*
 * Return RR DLZ BDBHPT rdata field. Fields are space (" ") delimited.
 */
static size_t edit_write_rr_field(struct iovec *iov, const unsigned char **pos, const unsigned char *end) {
	iov->iov_base	= 0;
	iov->iov_len	= 0;

	while (*pos < end && **pos == ' ')
		++*pos;

	if (*pos >= end)
		return 0;

	iov->iov_base	= (void *)*pos;

	while (*pos < end && **pos != ' ')
		++*pos;

	return iov->iov_len = (end - *pos);
} /* edit_write_rr_field() */


/*
 * Compare an RR and a DLZ BDBHPT rdata entity for equality. The RR can
 * contain wildcard values.
 *
 * FIXME: Broken for TXT RRs. We need to special case, and just memcmp()
 * from the beginning of the TXT string rdata field to the end of rdata
 * buffer.
 */
static int edit_write_rr_eq(const unsigned char *rdata, size_t rdlen, struct ZonedRR *rr, int exact) {
	struct { struct iovec fields[7]; } a = { { { 0, 0 } } }, b = { { { 0, 0 } } };
	unsigned char rdatab[256];
	const unsigned char *pos, *end;
	int i;

	pos	= rdata;
	end	= rdata + rdlen;

	for (i = 0; i < arraylen(a.fields); i++) {
		if (0 == edit_write_rr_field(&a.fields[i], &pos, end))
			break;
	}

	if (rr->present == ZonedRR_PR_all) {
		b.fields[1].iov_base	= rr->all.host.buf;
		b.fields[1].iov_len	= rr->all.host.size;

		b.fields[3].iov_base	= rr->all.rtype.buf;
		b.fields[3].iov_len	= rr->all.rtype.size;
	} else {
		rdlen	= edit_rdata(rdatab, sizeof rdatab, rr);

		if (rdlen == 0 || rdlen >= sizeof rdatab)
			return -1;

		pos	= rdatab;
		end	= rdatab + rdlen;

		for (i = 0; i < arraylen(b.fields); i++) {
			if (0 == edit_write_rr_field(&b.fields[i], &pos, end))
				break;
		}
	}

	/* .host */
	if (b.fields[1].iov_len == 0)	/* wildcard */
		return 1;
	if (a.fields[1].iov_len != b.fields[1].iov_len)
		return 0;
	if (0 != strncasecmp(a.fields[1].iov_base, b.fields[1].iov_base, a.fields[1].iov_len))
		return 0;

	/* .rtype */
	if (b.fields[3].iov_len == 0)	/* wildcard */
		return 1;
	if (a.fields[3].iov_len != b.fields[3].iov_len)
		return 0;
	if (0 != strncasecmp(a.fields[3].iov_base, b.fields[3].iov_base, a.fields[3].iov_len))
		return 0;

	if (!exact)
		return 1;

	for (i = 4; i < arraylen(a.fields); i++) {
		if (b.fields[i].iov_len == 0)	/* wildcard */
			return 1;
		if (a.fields[i].iov_len != b.fields[i].iov_len)
			return 0;
		if (0 != strncasecmp(a.fields[i].iov_base, b.fields[i].iov_base, a.fields[i].iov_len))
			return 0;
	}

	return 1;
} /* edit_write_rr_eq() */


static int edit_write_rr_set(struct zmbuf *mb, struct bdb *bdb, DB_TXN *txn, const char *logid, struct ZonedRR *rr) {
	DBC *cur	= 0;
	char keyb[256];
	unsigned char rdata[256];
	DBT key, data;
	int dberr, rval	= 0;

	key	= dbt_initializer;

	if (-1 == (key.size = snprintf(key.data = keyb, sizeof keyb, "%.*s %.*s", (int)mb->msg->edit.zone.size, (char *)mb->msg->edit.zone.buf, (int)rr->any.host.size, rr->any.host.buf)))
		return ENAMETOOLONG;

	if (0 != (dberr = bdb->db.data->cursor(bdb->db.data, txn, &cur, 0)))
		return dberr;

	data	= dbt_initializer;
	dberr	= cur->c_get(cur, &key, &data, DB_FIRST);

	/*
	 * FIXME: Handle DB_BUFFER_SMALL.
	 */
	while (dberr == 0) {
		if (edit_write_rr_eq(data.data, data.size, rr, 0)) {
			if (0 != (dberr = cur->c_del(cur, 0)) && dberr != DB_KEYEMPTY)
				break;
		}

		data	= dbt_initializer;
		dberr	= cur->c_get(cur, &key, &data, DB_NEXT_DUP);
	}

	if (dberr != DB_NOTFOUND) {
		rval	= dberr;

		goto epilog;
	}

	data	= dbt_initializer;

	if (0 != (data.size = edit_rdata(data.data = rdata, sizeof rdata, rr)) || data.size > sizeof rdata) {
		rval	= cur->c_put(cur, &key, &data, DB_KEYLAST);
	} else {
		warn("invalid RR");

		rval	= EINVAL;
	}
epilog:
	if (cur != 0 && 0 != (dberr = cur->c_close(cur)))
		warn("(%s) dbc_close: %s", logid, db_strerror(dberr));

	return rval;
} /* edit_write_rr_set() */


static int edit_write_rr_add(struct zmbuf *mb, struct bdb *bdb, DB_TXN *txn, const char *logid, struct ZonedRR *rr) {
	char keyb[256];
	unsigned char rdata[256];
	DBT key		= dbt_initializer;
	DBT data	= dbt_initializer;

	if (-1 == (key.size = snprintf(key.data = keyb, sizeof keyb, "%.*s %.*s", (int)mb->msg->edit.zone.size, (char *)mb->msg->edit.zone.buf, (int)rr->any.host.size, rr->any.host.buf)))
		return ENAMETOOLONG;

	if (0 == (data.size = edit_rdata(data.data = rdata, sizeof rdata, rr)) || data.size > sizeof rdata) {
		warn("invalid RR");

		return EINVAL;
	}

	return bdb->db.data->put(bdb->db.data, txn, &key, &data, 0);
} /* edit_write_rr_add() */


static int edit_write_rr_del(struct zmbuf *mb, struct bdb *bdb, DB_TXN *txn, const char *logid, struct ZonedRR *rr) {
	DBC *cur	= 0;
	char keyb[256];
	DBT key, data;
	int dberr, rval	= 0;

	key	= dbt_initializer;

	if (-1 == (key.size = snprintf(key.data = keyb, sizeof keyb, "%.*s %.*s", (int)mb->msg->edit.zone.size, (char *)mb->msg->edit.zone.buf, (int)rr->any.host.size, rr->any.host.buf)))
		return ENAMETOOLONG;

	if (0 != (dberr = bdb->db.data->cursor(bdb->db.data, txn, &cur, 0)))
		return dberr;

	data	= dbt_initializer;
	dberr	= cur->c_get(cur, &key, &data, DB_FIRST);

	while (dberr == 0) {
		if (edit_write_rr_eq(data.data, data.size, rr, 1)) {
			if (0 != (dberr = cur->c_del(cur, 0)) && dberr != DB_KEYEMPTY)
				break;
		}

		data	= dbt_initializer;
		dberr	= cur->c_get(cur, &key, &data, DB_NEXT_DUP);
	}

	if (dberr != DB_NOTFOUND) {
		rval	= dberr;

		goto epilog;
	}

	rval	= 0;
epilog:
	if (cur != 0 && 0 != (dberr = cur->c_close(cur)))
		warn("(%s) dbc_close: %s", logid, db_strerror(dberr));

	return rval;
} /* edit_write_rr_del() */


static const char *edit_write_strmode(enum ZonedEditMode mode) {
	switch (mode) {
	case ZonedEditMode_editModeSet:
		return "set";
	case ZonedEditMode_editModeAdd:
		return "add";
	case ZonedEditMode_editModeDel:
		return "del";
	default:
		return "?";
	}

	/* NOT REACHED */
} /* edit_write_strmode() */


static enum { EDIT_FAIL = -1, EDIT_RETRY = 0, EDIT_OKAY = 1 } edit_write(struct bdb *bdb, struct zmbuf *mb) {
	static const typeof(&edit_write_zone_set) edit[2][3] = {
		{ &edit_write_zone_set, &edit_write_zone_set, &edit_write_zone_del },
		{ &edit_write_rr_set, &edit_write_rr_add, &edit_write_rr_del },
	};
	char logid[256];
	char *zone	= (char *)mb->msg->edit.zone.buf;
	size_t zlen	= mb->msg->edit.zone.size;
	int tryno	= 0;
	long l;
	int scope, mode;
	DB_TXN *txn	= 0;
	struct ZonedRR *rr;
	int rrno;
	int dberr, rval	= EDIT_FAIL;

	assert(0 < snprintf(logid, sizeof logid, "%.*s/?", (int)MIN(zlen, sizeof logid - 16), zone));

	scope		= !!mb->msg->edit.rrset.list.count;

	if (0 != asn_INTEGER2long(&mb->msg->edit.mode, &l)) {
		error("(%s) unable to determine mode", logid);

		return EDIT_FAIL;
	}

	switch (l) {
	case ZonedEditMode_editModeSet:
		mode	= 0; break;
	case ZonedEditMode_editModeAdd:
		mode	= 1; break;
	case ZonedEditMode_editModeDel:
		mode	= 2; break;
	default:
		error("(%s) unknown edit mode: %ld", logid, l);

		return EDIT_FAIL;
	}

	assert(0 < snprintf(logid, sizeof logid, "%.*s/%s", (int)MIN(zlen, sizeof logid - 16), zone, edit_write_strmode(l)));

begin:
	if (txn != 0 && 0 != (dberr = txn->abort(txn)))
		panic("(%s) failed to abort transaction: %s", logid, db_strerror(dberr));

	txn	= 0;

	if (++tryno > 7) {
		rval	= EDIT_RETRY;

		goto epilog;
	}

	if (0 != (dberr = bdb->db.env->txn_begin(bdb->db.env, 0, &txn, DB_TXN_SNAPSHOT))) {
		error("(%s) unable to open transaction: %s", logid, db_strerror(dberr));

		return -1;
	}

	rrno	= 0;

	do {
		if (rrno < mb->msg->edit.rrset.list.count)
			rr	= mb->msg->edit.rrset.list.array[rrno];
		else
			rr	= 0;

		dberr	= edit[scope][mode](mb, bdb, txn, logid, rr);

		if (dberr != 0)
			debug0("(%s) edit: %s", logid, db_strerror(dberr));

		switch (dberr) {
		case 0:
			if (0 != (dberr = txn->commit(txn, 0)))
				error("(%s) failed to commit transaction: %s", logid, db_strerror(dberr));
			else
				rval	= EDIT_OKAY;

			txn	= 0;

			break;
		case DB_LOCK_DEADLOCK:
			goto begin;
		case DB_REP_HANDLE_DEAD:
			bdb_openall(bdb, 1);

			goto begin;
		case DB_REP_LOCKOUT:
			rval	= EDIT_RETRY;
			
			goto epilog;
		default:
			error("(%s) unknown DB error: %s", logid, db_strerror(dberr));

			goto epilog;
		} /* switch (dberr) */
	} while (++rrno < mb->msg->edit.rrset.list.count);
epilog:
	if (txn != 0 && 0 != (dberr = txn->abort(txn)))
		panic("(%s) failed to abort transaction: %s", logid, db_strerror(dberr));

	return rval;
} /* edit_write() */


static void *edit_writer(void *arg0, void *arg1, unsigned arg2) {
	struct zmbuf *mb0, *mb;
	int masterid;

	pthread_mutex_lock(&edits.mutex);

	for (;;) {
		if (0 < (masterid = DNSDB.master.envid) && SIMPLEQ_END(&edits.queue) != (mb0 = SIMPLEQ_FIRST(&edits.queue))) {
			do {
				if (SIMPLEQ_END(&edits.queue) == (mb = SIMPLEQ_FIRST(&edits.queue)))
					break;

				SIMPLEQ_REMOVE_HEAD(&edits.queue, sqe);

				/*
				 * Unlock so others can call edit_apply(). 
				 * Otherwise there may be a deadlock lurking
				 * betwixt edit_apply() and zmbuf_send().
				 */
				pthread_mutex_unlock(&edits.mutex);

				if (masterid == BDB_SELF_EID) {
					if (EDIT_RETRY != edit_write(&DNSDB, mb))
						zmbuf_free(mb), mb = 0;
				} else {
					if (0 == zmbuf_send(mb, masterid, 0))
						mb	= 0;
				}

				pthread_mutex_lock(&edits.mutex);

				if (mb != 0)
					SIMPLEQ_INSERT_TAIL(&edits.queue, mb, sqe);
			} while (mb0 != mb);
		}

		if (0 != pthread_cond_wait(&edits.gate, &edits.mutex))
			panic("pthread_cond_wait: %s", strerror(errno));
	}

	pthread_mutex_unlock(&edits.mutex);

	return 0;
} /* edit_writer() */


void edit_apply(struct zmbuf *mb) {
	(void)pthread_mutex_lock(&edits.mutex);

	SIMPLEQ_INSERT_TAIL(&edits.queue, mb, sqe);

	(void)pthread_cond_signal(&edits.gate);

	(void)pthread_mutex_unlock(&edits.mutex);

	return /* void */;
} /* edit_apply() */


void edit_wake(void) {
	(void)pthread_mutex_lock(&edits.mutex);

	(void)pthread_cond_signal(&edits.gate);

	(void)pthread_mutex_unlock(&edits.mutex);

	return /* void */;
} /* edit_wake() */


void edit_init(void) {
	thread_run(&edit_writer, 0, THREAD_PRI_NORM, 0, 0, 0);

	return /* void */;
} /* edit_init() */

